"""Tests for metabot.modules.reminders."""

import datetime

import pytest
import pytz
import yaml

from metabot.calendars import loader
from metabot.modules import events
from metabot.modules import moderator
from metabot.modules import reminders

# pylint: disable=too-many-lines


@pytest.fixture
def conversation(build_conversation, monkeypatch):  # pylint: disable=missing-docstring
    monkeypatch.setattr('time.time', lambda: 2000.)

    cal = loader.get('static:test_events')
    assert cal.calcode == '6fc2c510'
    cal.events = {
        '6fc2c510:alpha': {
            'description': 'Alpha Description',
            'end': 60 * 60 * 3.5,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 60 * 60 * 2.5,
            'summary': 'Alpha Summary',
        },
        '6fc2c510:bravo': {
            'description': 'Bravo Description',
            'end': 60 * 60 * 24 * 7 + 60 * 60,
            'local_id': '6fc2c510:bravo',
            'location': 'Bravo Venue, Rest of Bravo Location',
            'start': 60 * 60 * 24 * 7,
            'summary': 'Bravo Summary',
        },
        '6fc2c510:charlie': {
            'description': 'Charlie Description',
            'end': 60 * 60 * 24 * 7 + 60 + 60 * 60,
            'local_id': '6fc2c510:charlie',
            'location': 'Charlie Venue, Rest of Charlie Location',
            'start': 60 * 60 * 24 * 7 + 60,
            'summary': 'Charlie Summary',
        },
    }

    conv = build_conversation(events, moderator)
    conv.multibot.multical.add('static:test_events')
    conv.multibot.calendars['6fc2c510'] = {'name': 'Test Calendar'}
    return conv


# pylint: disable=line-too-long


def test_generate_preamble():  # pylint: disable=missing-docstring
    # yapf: disable - pylint: disable=protected-access
    assert reminders._generate_preamble('', []) == (
        'No upcoming events!')
    assert reminders._generate_preamble('1 + 1 = 2!', []) == (
        '1 + 1 = 2!')

    assert reminders._generate_preamble('', ['EVENT1']) == (
        "There's an event coming up:")
    assert reminders._generate_preamble('', ['EVENT1', 'EVENT2']) == (
        "There are a couple events coming up:")
    assert reminders._generate_preamble('', ['EVENT1', 'EVENT2', 'EVENT3']) == (
        "There are a few events coming up:")
    assert reminders._generate_preamble('', ['EVENT1', 'EVENT2', 'EVENT3', 'EVENT4']) == (
        "There are a bunch of events coming up:")
    assert reminders._generate_preamble('', ['EVENT1', 'EVENT2', 'EVENT3', 'EVENT4', 'EVENT5']) == (
        "There are a bunch of events coming up:")

    assert reminders._generate_preamble('1 + 1 = 2!', ['EVENT1']) == (
        "1 + 1 = 2! Also, there's an event coming up:")

    assert reminders._generate_preamble('Visit @MYGROUP', ['EVENT1']) == (
        "Visit @MYGROUP • Also, there's an event coming up:")

    assert reminders._generate_preamble('I love events.', ['EVENT1']) == (
        "I love events. Speaking of which, there's an event coming up:")

    assert reminders._generate_preamble('I love events:', ['EVENT1']) == (
        "I love events:")
    # yapf: enable


@pytest.fixture
def daily_messages(conversation):  # pylint: disable=missing-docstring,redefined-outer-name
    replies = conversation.raw_message('/dummy')
    assert conversation.format_messages(replies) == ''

    records = {}

    def _daily_messages(initial=False):
        replies.clear()
        if initial:
            records.clear()
            conversation.bot.config['issue37']['moderator']['-1002000002000']['daily']['hour'] = 0
        else:
            conversation.bot.config['issue37']['moderator']['-1002000002000']['daily']['hour'] = 1
        reminders._daily_messages(conversation.multibot, records)  # pylint: disable=protected-access
        outrecords = {'/'.join(key): record for key, record in records.items()}
        text = outrecords and yaml.safe_dump(outrecords, width=float('inf'),
                                             allow_unicode=True) or ''
        text = text.replace('\n\n', '\n')
        text = f'{text}\n\n{conversation.format_messages(replies)}'.strip()
        if text:
            text = f'\n{text}\n'
        return text

    assert _daily_messages(True) == ''

    conversation.bot.config['issue37']['moderator']['-1002000002000'] = {
        'calendars': '6fc2c510',
        'daily': {
            'hour': 0,
        },
        'timezone': 'UTC',
    }

    return _daily_messages


def test_daily_messages(daily_messages):  # pylint: disable=redefined-outer-name
    """Test basic daily announcement logic (essentially, test the daily_messages fixture)."""

    assert daily_messages(True) == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''
"""


def test_daily_messages_updated(daily_messages):  # pylint: disable=redefined-outer-name
    """Test daily announcement update messages."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['summary'] = 'Edited Summary'
    cal.events['6fc2c510:alpha']['end'] += 60

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12660.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Edited Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Edited Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:31ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
• <s>Alpha Summary</s> is now called <b>Edited Summary</b> and was extended to <b>Thu 1ˢᵗ, 3:31ᵃᵐ</b> (instead of <s>Thu 1ˢᵗ, 3:30ᵃᵐ</s>).


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Edited Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:31ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>]
"""

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12660.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Edited Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Edited Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:31ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>
"""


def test_daily_messages_future(daily_messages, monkeypatch):  # pylint: disable=redefined-outer-name
    """Test the behavior when an event transitions from TODAY to NOW."""

    assert daily_messages(True) == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    # See https://github.com/nmlorg/metabot/issues/97.
    monkeypatch.setattr('time.time', lambda: 60 * 60 * 2.5)

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">⭐ ᴺᴼᵂ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">⭐ ᴺᴼᵂ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['summary'] = 'Now Summary'

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Now Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Now Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">⭐ ᴺᴼᵂ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 2:30ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
• <s>Alpha Summary</s> is now called <b>Now Summary</b>.


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Now Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">⭐ ᴺᴼᵂ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12346">Updated 2:30ᵃᵐ</a>]
"""


def test_daily_messages_multiline(daily_messages):  # pylint: disable=redefined-outer-name
    """Verify multiline event components are rendered in updates concisely."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['description'] = 'Multi\n\nLine\nDescription'

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: 'Multi

      Line
      Description'
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
• The description of <b>Alpha Summary</b> was changed from <s>Alpha Description</s> to <b>Multi Line Description</b>.


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>]
"""


def test_daily_messages_add_remove_event(conversation, daily_messages):  # pylint: disable=redefined-outer-name
    """Verify what happens when events are added/removed entirely."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:new'] = {
        'description': 'New Description',
        'end': 60 * 60 * 4.5,
        'local_id': '6fc2c510:new',
        'location': 'New Venue, Rest of New Location',
        'start': 60 * 60 * 3.5,
        'summary': 'New Summary',
    }
    conversation.multibot.multical.ordered.append(cal.events['6fc2c510:new'])
    cal.events.pop('6fc2c510:bravo')
    assert conversation.multibot.multical.ordered[1]['local_id'] == '6fc2c510:bravo'
    assert conversation.multibot.multical.ordered.pop(1)
    conversation.multibot.multical._rebuild()  # pylint: disable=protected-access

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: New Description
    end: 16200.0
    local_id: 6fc2c510:new
    location: New Venue, Rest of New Location
    start: 12600.0
    summary: New Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>New Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpuZXcgVVRD">🔜 ³ ʰᵒᵘʳˢ Thu 1ˢᵗ, 3:30–4:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=New+Venue%2C+Rest+of+New+Location">New Venue</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
• <b>New Summary</b> was added.
• <s>Bravo Summary</s> was removed.


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>New Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpuZXcgVVRD">🔜 ³ ʰᵒᵘʳˢ Thu 1ˢᵗ, 3:30–4:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=New+Venue%2C+Rest+of+New+Location">New Venue</a>

[<a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>]
"""


def test_daily_messages_ignored(daily_messages):  # pylint: disable=redefined-outer-name
    """Verify non-notification-worthy changes are properly ignored."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['description'] = (
        'Alpha Description https://www.example.com/?id=1234')

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''
"""

    cal.events['6fc2c510:alpha']['updated'] = 23456

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''
"""


def test_daily_messages_icons(conversation, daily_messages):  # pylint: disable=redefined-outer-name
    """Run through all the cases where events trigger banner images."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['description'] = 'Board Games!'

    # An update does not include references to icons.
    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Board Games!
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
• The description of <b>Alpha Summary</b> was changed from <s>Alpha Description</s> to <b>Board Games!</b>.


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>]
"""

    # But an initial announcement does (photo=...).
    assert daily_messages(True) == """
modulestestbot/-1002000002000:
- 1800
- - description: Board Games!
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- caption: CAPTION
  chat:
    id: -1002000002000
  message_id: 12347
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''


[send_photo chat_id=-1002000002000 disable_notification=True parse_mode=HTML photo=https://ssl.gstatic.com/calendar/images/eventillustrations/v1/img_gamenight_2x.jpg]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    conversation.bot.config['issue37']['events']['series']['ha sum'] = 'SERIES ICON'

    assert daily_messages(True) == """
modulestestbot/-1002000002000:
- 1800
- - description: Board Games!
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- caption: CAPTION
  chat:
    id: -1002000002000
  message_id: 12348
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''


[send_photo chat_id=-1002000002000 disable_notification=True parse_mode=HTML photo=SERIES ICON]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    conversation.bot.config['issue37']['events']['events']['6fc2c510:alpha'] = 'EVENT ICON'

    assert daily_messages(True) == """
modulestestbot/-1002000002000:
- 1800
- - description: Board Games!
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- caption: CAPTION
  chat:
    id: -1002000002000
  message_id: 12349
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''


[send_photo chat_id=-1002000002000 disable_notification=True parse_mode=HTML photo=EVENT ICON]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    cal.events['6fc2c510:alpha']['description'] = 'Fun Games!'

    # Removing an icon trigger leaves the image in place.
    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Fun Games!
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- caption: CAPTION
  chat:
    id: -1002000002000
  message_id: 12349
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- <a href="https://t.me/c/2000002000/12350">Updated 12:33ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12349]
Updated:
• The description of <b>Alpha Summary</b> was changed from <s>Board Games!</s> to <b>Fun Games!</b>.


[edit_message_caption chat_id=-1002000002000 message_id=12349 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12350">Updated 12:33ᵃᵐ</a>]
"""


def test_daily_messages_geometry(conversation, daily_messages):  # pylint: disable=redefined-outer-name
    """See https://github.com/nmlorg/metabot/issues/75."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['description'] = 'Trigger'
    conversation.bot.config['issue37']['moderator']['-1002000002000']['maxeventscount'] = 1

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Trigger
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There''s an event coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
• The description of <b>Alpha Summary</b> was changed from <s>Alpha Description</s> to <b>Trigger</b>.


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There's an event coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>

[<a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>]
"""


def test_daily_messages_grammar(daily_messages):  # pylint: disable=redefined-outer-name
    """Quick tests for the differ's grammar."""

    daily_messages(True)
    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['summary'] = 'New Summary'
    cal.events['6fc2c510:alpha']['start'] -= 1800
    cal.events['6fc2c510:alpha']['location'] = 'New Location'
    cal.events['6fc2c510:alpha']['description'] = 'New Description'
    cal.events['6fc2c510:bravo']['start'] -= 1800
    cal.events['6fc2c510:bravo']['end'] += 1800
    cal.events['6fc2c510:bravo']['location'] = 'New Location'
    cal.events['6fc2c510:bravo']['description'] = 'New Description'

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: New Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: New Location
    start: 7200.0
    summary: New Summary
  - description: New Description
    end: 610200
    local_id: 6fc2c510:bravo
    location: New Location
    start: 603000
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>New Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ¹ʰ³⁰ᵐ Thu 1ˢᵗ, 2–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=New+Location">New Location</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">⁶ ᵈᵃʸˢ Wed 7ᵗʰ, 11:30ᵖᵐ – 1:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=New+Location">New Location</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
• <s>Alpha Summary</s> is now called <b>New Summary</b>, is starting earlier at <b>Thu 1ˢᵗ, 2ᵃᵐ</b> (instead of <s>Thu 1ˢᵗ, 2:30ᵃᵐ</s>; same end), was moved from <s>Alpha Venue</s> to <b>New Location</b>, and its description was changed from <s>Alpha Description</s> to <b>New Description</b>.
• <b>Bravo Summary</b> now starts at <b>Wed 7ᵗʰ, 11:30ᵖᵐ</b> and ends at <b>Thu 8ᵗʰ, 1:30ᵃᵐ</b> (was <s>Thu 8ᵗʰ, 12ᵃᵐ to Thu 8ᵗʰ, 1ᵃᵐ</s>), was moved from <s>Bravo Venue</s> to <b>New Location</b>, and its description was changed from <s>Bravo Description</s> to <b>New Description</b>.


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>New Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ¹ʰ³⁰ᵐ Thu 1ˢᵗ, 2–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=New+Location">New Location</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">⁶ ᵈᵃʸˢ Wed 7ᵗʰ, 11:30ᵖᵐ – 1:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=New+Location">New Location</a>

[<a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>]
"""


def test_daily_messages_replaced(conversation, daily_messages, monkeypatch):  # pylint: disable=redefined-outer-name
    """Test what happens when an existing announcement is obsoleted the next day."""

    monkeypatch.setattr('time.time', lambda: 4000)

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 3600
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ¹ʰ³⁰ᵐ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ¹ʰ³⁰ᵐ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    monkeypatch.setattr('time.time', lambda: 4000 + 60 * 60 * 24)

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 90000
- - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
  - description: Charlie Description
    end: 608460
    local_id: 6fc2c510:charlie
    location: Charlie Venue, Rest of Charlie Location
    start: 604860
    summary: Charlie Summary
- chat:
    id: -1002000002000
  message_id: 12346
- 'There are a couple events coming up:

  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">⁶ ᵈᵃʸˢ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
  <b>Charlie Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVUQw">⁶ ᵈᵃʸˢ Thu 8ᵗʰ, 12:01–1:01ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>'
- ''


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
There are a couple events coming up:

<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">⁶ ᵈᵃʸˢ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
<b>Charlie Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVUQw">⁶ ᵈᵃʸˢ Thu 8ᵗʰ, 12:01–1:01ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    conversation.bot.config['issue37']['moderator']['-1002000002000']['daily']['pin'] = True
    monkeypatch.setattr('time.time', lambda: 4000 + 60 * 60 * 24 * 2)

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 176400
- - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
  - description: Charlie Description
    end: 608460
    local_id: 6fc2c510:charlie
    location: Charlie Venue, Rest of Charlie Location
    start: 604860
    summary: Charlie Summary
- chat:
    id: -1002000002000
  message_id: 12347
  pinned: true
- 'There are a couple events coming up:

  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">⁵ ᵈᵃʸˢ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
  <b>Charlie Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVUQw">⁵ ᵈᵃʸˢ Thu 8ᵗʰ, 12:01–1:01ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>'
- ''


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
There are a couple events coming up:

<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">⁵ ᵈᵃʸˢ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
<b>Charlie Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVUQw">⁵ ᵈᵃʸˢ Thu 8ᵗʰ, 12:01–1:01ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>


[pin_chat_message chat_id=-1002000002000 disable_notification=True message_id=12347]
(EMPTY MESSAGE)


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12346 parse_mode=HTML]
There are a couple events coming up:

<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
<b>Charlie Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVUQw">Thu 8ᵗʰ, 12:01–1:01ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>
"""

    monkeypatch.setattr('time.time', lambda: 4000 + 60 * 60 * 24 * 3)

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 262800
- - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
  - description: Charlie Description
    end: 608460
    local_id: 6fc2c510:charlie
    location: Charlie Venue, Rest of Charlie Location
    start: 604860
    summary: Charlie Summary
- chat:
    id: -1002000002000
  message_id: 12349
  pinned: true
- 'There are a couple events coming up:

  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">⁴ ᵈᵃʸˢ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
  <b>Charlie Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVUQw">⁴ ᵈᵃʸˢ Thu 8ᵗʰ, 12:01–1:01ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>'
- ''


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
There are a couple events coming up:

<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">⁴ ᵈᵃʸˢ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
<b>Charlie Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVUQw">⁴ ᵈᵃʸˢ Thu 8ᵗʰ, 12:01–1:01ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>


[pin_chat_message chat_id=-1002000002000 disable_notification=True message_id=12349]
(EMPTY MESSAGE)


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12347 parse_mode=HTML]
There are a couple events coming up:

<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
<b>Charlie Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVUQw">Thu 8ᵗʰ, 12:01–1:01ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>


[unpin_chat_message chat_id=-1002000002000 message_id=12347]
(EMPTY MESSAGE)
"""


def test_diff_time():
    """Test start/end time difference detector."""

    tzinfo = pytz.timezone('America/Los_Angeles')
    base = datetime.datetime.fromtimestamp(1719946200, tzinfo)
    start = 1719946800.
    end = 1719950400

    # pylint: disable=protected-access

    assert not reminders._diff_time(start, end, start, end, tzinfo, base)

    assert reminders._diff_time(start, end, start + 60, end + 120, tzinfo, base) == (
        'now starts at <b>Tue 2ⁿᵈ, 12:01ᵖᵐ</b> and ends at <b>Tue 2ⁿᵈ, 1:02ᵖᵐ</b> (was <s>Tue 2ⁿᵈ, 12ᵖᵐ to Tue 2ⁿᵈ, 1ᵖᵐ</s>)'
    )

    assert reminders._diff_time(start, end, start - 60, end, tzinfo, base) == (
        'is starting earlier at <b>Tue 2ⁿᵈ, 11:59ᵃᵐ</b> (instead of <s>Tue 2ⁿᵈ, 12ᵖᵐ</s>; same end)'
    )
    assert reminders._diff_time(start, end, start - 60, end - 60, tzinfo, base) == (
        'was moved up to <b>Tue 2ⁿᵈ, 11:59ᵃᵐ</b> (from <s>Tue 2ⁿᵈ, 12ᵖᵐ</s>; same duration)')

    assert reminders._diff_time(start, end, start + 60, end, tzinfo, base) == (
        'is starting later at <b>Tue 2ⁿᵈ, 12:01ᵖᵐ</b> (instead of <s>Tue 2ⁿᵈ, 12ᵖᵐ</s>; same end)')
    assert reminders._diff_time(start, end, start + 60, end + 60, tzinfo, base) == (
        'was moved back to <b>Tue 2ⁿᵈ, 12:01ᵖᵐ</b> (from <s>Tue 2ⁿᵈ, 12ᵖᵐ</s>; same duration)')

    assert reminders._diff_time(
        start, end, start, end - 60, tzinfo,
        base) == ('was shortened to <b>Tue 2ⁿᵈ, 12:59ᵖᵐ</b> (instead of <s>Tue 2ⁿᵈ, 1ᵖᵐ</s>)')
    assert reminders._diff_time(
        start, end, start, end + 60, tzinfo,
        base) == ('was extended to <b>Tue 2ⁿᵈ, 1:01ᵖᵐ</b> (instead of <s>Tue 2ⁿᵈ, 1ᵖᵐ</s>)')


def test_quick_diff():
    """Test description string differ."""

    # pylint: disable=protected-access
    assert reminders._quick_diff('', '') is None
    assert reminders._quick_diff('0987654321', '0987654321new') == ('0987654321', '0987654321new')
    assert reminders._quick_diff('10987654321', '10987654321new') == ('…987654321', '…987654321new')
    assert reminders._quick_diff('10987654321old', '10987654321') == ('…987654321old', '…987654321')
    assert reminders._quick_diff('10987654321old',
                                 '10987654321new') == ('…987654321old', '…987654321new')
    assert reminders._quick_diff(
        'x',
        'x234567890123456789012345678901234567890') == ('x',
                                                        'x234567890123456789012345678901234567890')
    assert reminders._quick_diff(
        'x',
        'x2345678901234567890123456789012345678901') == ('x',
                                                         'x23456789012345678901234567890123456789…')


def test_truncation(daily_messages, monkeypatch):  # pylint: disable=redefined-outer-name
    """Verify that message limits are correctly enforced."""

    monkeypatch.setattr('ntelebot.limits.message_text_length_max', 40)

    assert daily_messages(True) == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Alpha Venue, Rest of Alpha Location
    start: 9000.0
    summary: Alpha Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Alpha Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- ''


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML]
There are a couple events coming up:

<b>Al</b>
"""
    assert len('There are a couple events coming up:\n\nAl') == 40


def test_sanitize(daily_messages):  # pylint: disable=redefined-outer-name
    """Verify entity escaping."""

    daily_messages(True)

    cal = loader.get('static:test_events')
    cal.events['6fc2c510:alpha']['summary'] = 'Edited & <i> Summary'
    cal.events['6fc2c510:alpha']['location'] = 'Edited & <i> Location, Rest of Location'

    assert daily_messages() == """
modulestestbot/-1002000002000:
- 1800
- - description: Alpha Description
    end: 12600.0
    local_id: 6fc2c510:alpha
    location: Edited & <i> Location, Rest of Location
    start: 9000.0
    summary: Edited & <i> Summary
  - description: Bravo Description
    end: 608400
    local_id: 6fc2c510:bravo
    location: Bravo Venue, Rest of Bravo Location
    start: 604800
    summary: Bravo Summary
- chat:
    id: -1002000002000
  message_id: 12345
- 'There are a couple events coming up:

  <b>Edited &amp; &lt;i&gt; Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Edited+%26+%3Ci%3E+Location%2C+Rest+of+Location">Edited &amp; &lt;i&gt; Location</a>
  <b>Bravo Summary</b>
  <a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>'
- <a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>


[chat_id=-1002000002000 disable_notification=True disable_web_page_preview=True parse_mode=HTML reply_to_message_id=12345]
Updated:
• <s>Alpha Summary</s> is now called <b>Edited &amp; &lt;i&gt; Summary</b> and was moved from <s>Alpha Venue</s> to <b>Edited &amp; &lt;i&gt; Location</b>.


[edit_message_text chat_id=-1002000002000 disable_web_page_preview=True message_id=12345 parse_mode=HTML]
There are a couple events coming up:

<b>Edited &amp; &lt;i&gt; Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">🔜 ² ʰᵒᵘʳˢ Thu 1ˢᵗ, 2:30–3:30ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Edited+%26+%3Ci%3E+Location%2C+Rest+of+Location">Edited &amp; &lt;i&gt; Location</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

[<a href="https://t.me/c/2000002000/12346">Updated 12:33ᵃᵐ</a>]
"""



"""Tests for metabot.util.humanize."""

import datetime

from metabot.util import humanize


def test_humanize_date():
    """Quick tests for humanize.date."""

    base = datetime.date(2017, 11, 15)
    assert humanize.date(datetime.date(2017, 11, 15), base=base) == 'Wed 15ᵗʰ'
    assert humanize.date(datetime.date(2017, 11, 16), base=base) == 'Thu 16ᵗʰ'


def test_humanize_dayofmonth():
    """Quick tests for humanize.dayofmonth."""

    assert humanize.dayofmonth(1) == '1ˢᵗ'
    assert humanize.dayofmonth(2) == '2ⁿᵈ'
    assert humanize.dayofmonth(3) == '3ʳᵈ'
    assert humanize.dayofmonth(4) == '4ᵗʰ'
    assert humanize.dayofmonth(11) == '11ᵗʰ'
    assert humanize.dayofmonth(12) == '12ᵗʰ'
    assert humanize.dayofmonth(13) == '13ᵗʰ'
    assert humanize.dayofmonth(21) == '21ˢᵗ'


def test_nextmonth():
    """Quick tests for humanize._nextmonth."""

    # pylint: disable=protected-access
    assert humanize._nextmonth(datetime.date(2024, 11, 1), 1) == datetime.date(2024, 12, 1)
    assert humanize._nextmonth(datetime.date(2024, 12, 1), -1) == datetime.date(2024, 11, 1)
    assert humanize._nextmonth(datetime.date(2024, 12, 1), 1) == datetime.date(2025, 1, 1)
    assert humanize._nextmonth(datetime.date(2025, 1, 1), -1) == datetime.date(2024, 12, 1)

    assert humanize._nextmonth(datetime.date(2025, 1, 31), 1) == datetime.date(2025, 2, 28)
    assert humanize._nextmonth(datetime.date(2025, 1, 31), 2) == datetime.date(2025, 3, 31)
    assert humanize._nextmonth(datetime.date(2025, 1, 31), 3) == datetime.date(2025, 4, 30)
    assert humanize._nextmonth(datetime.date(2025, 1, 31), 4) == datetime.date(2025, 5, 31)

    assert humanize._nextmonth(datetime.date(2025, 5, 31), -1) == datetime.date(2025, 4, 30)
    assert humanize._nextmonth(datetime.date(2025, 5, 31), -2) == datetime.date(2025, 3, 31)
    assert humanize._nextmonth(datetime.date(2025, 5, 31), -3) == datetime.date(2025, 2, 28)
    assert humanize._nextmonth(datetime.date(2025, 5, 31), -4) == datetime.date(2025, 1, 31)


def test_nextyear():
    """Quick tests for humanize._nextyear."""

    # pylint: disable=protected-access
    assert humanize._nextyear(datetime.date(2024, 11, 1), 1) == datetime.date(2025, 11, 1)
    assert humanize._nextyear(datetime.date(2025, 11, 1), -1) == datetime.date(2024, 11, 1)

    assert humanize._nextyear(datetime.date(2024, 2, 29), 1) == datetime.date(2025, 2, 28)
    assert humanize._nextyear(datetime.date(2024, 2, 29), 2) == datetime.date(2026, 2, 28)
    assert humanize._nextyear(datetime.date(2024, 2, 29), 3) == datetime.date(2027, 2, 28)
    assert humanize._nextyear(datetime.date(2024, 2, 29), 4) == datetime.date(2028, 2, 29)

    assert humanize._nextyear(datetime.date(2028, 2, 29), -1) == datetime.date(2027, 2, 28)
    assert humanize._nextyear(datetime.date(2028, 2, 29), -2) == datetime.date(2026, 2, 28)
    assert humanize._nextyear(datetime.date(2028, 2, 29), -3) == datetime.date(2025, 2, 28)
    assert humanize._nextyear(datetime.date(2028, 2, 29), -4) == datetime.date(2024, 2, 29)


def test_humanize_howrecent():
    """Quick tests for humanize.howrecent."""

    def _test(*spec):
        return humanize.howrecent(datetime.datetime(2017, 11, 15, 12),
                                  datetime.datetime(2017, 11, 15, 13),
                                  base=datetime.datetime(*spec))

    assert _test(2015, 10, 15) == '²ʸ¹ᵐ'
    assert _test(2015, 10, 16) == '²ʸ'
    assert _test(2015, 11, 15) == '²ʸ'
    assert _test(2016, 11, 15) == '¹ʸ'
    assert _test(2016, 11, 16) == '¹¹ᵐ⁴ʷ'
    assert _test(2017, 9, 15) == '²ᵐ'
    assert _test(2017, 10, 8) == '¹ᵐ¹ʷ'
    assert _test(2017, 10, 9) == '¹ᵐ'
    assert _test(2017, 10, 15) == '¹ᵐ'
    assert _test(2017, 10, 16) == '⁴ʷ²ᵈ'
    assert _test(2017, 11, 1) == '²ʷ'
    assert _test(2017, 11, 2) == '¹ʷ⁶ᵈ'
    assert _test(2017, 11, 8) == '¹ʷ'
    assert _test(2017, 11, 9) == '⁶ ᵈᵃʸˢ'
    assert _test(2017, 11, 13) == '² ᵈᵃʸˢ'
    assert _test(2017, 11, 14) == 'ᵗᵒᵐᵒʳʳᵒʷ'
    assert _test(2017, 11, 14, 12) == '🔜 ²⁴ ʰᵒᵘʳˢ'
    assert _test(2017, 11, 14, 23, 50) == '🔜 ¹²ʰ¹⁰ᵐ'

    assert _test(2017, 11, 15) == '🔜 ¹² ʰᵒᵘʳˢ'
    assert _test(2017, 11, 15, 9) == '🔜 ³ ʰᵒᵘʳˢ'
    assert _test(2017, 11, 15, 9, 30) == '🔜 ²ʰ³⁰ᵐ'
    assert _test(2017, 11, 15, 10) == '🔜 ² ʰᵒᵘʳˢ'
    assert _test(2017, 11, 15, 10, 30) == '🔜 ¹ʰ³⁰ᵐ'
    assert _test(2017, 11, 15, 10, 50) == '🔜 ¹ʰ¹⁰ᵐ'
    assert _test(2017, 11, 15, 11) == '🔜 ¹ ʰᵒᵘʳ'
    assert _test(2017, 11, 15, 11, 30) == '🔜 ³⁰ ᵐⁱⁿ'
    assert _test(2017, 11, 15, 11, 50) == '🔜 ¹⁰ ᵐⁱⁿ'
    assert _test(2017, 11, 15, 12) == '⭐ ᴺᴼᵂ'
    assert _test(2017, 11, 15, 12, 30) == '⭐ ᴺᴼᵂ'
    assert _test(2017, 11, 15, 13) == '⭐ ᴺᴼᵂ'
    assert _test(2017, 11, 15, 13, 10) == 'ᵗᵒᵈᵃʸ'
    assert _test(2017, 11, 15, 14) == 'ᵗᵒᵈᵃʸ'

    assert _test(2017, 11, 16) == 'ʸᵉˢᵗᵉʳᵈᵃʸ'
    assert _test(2017, 11, 17) == '² ᵈᵃʸˢ ᵃᵍᵒ'
    assert _test(2017, 11, 21) == '⁶ ᵈᵃʸˢ ᵃᵍᵒ'
    assert _test(2017, 11, 22) == '¹ʷ ᵃᵍᵒ'
    assert _test(2017, 11, 28) == '¹ʷ⁶ᵈ ᵃᵍᵒ'
    assert _test(2017, 11, 29) == '²ʷ ᵃᵍᵒ'
    assert _test(2018, 11, 14) == '¹¹ᵐ⁴ʷ ᵃᵍᵒ'
    assert _test(2018, 11, 15) == '¹ʸ ᵃᵍᵒ'
    assert _test(2018, 12, 15) == '¹ʸ¹ᵐ ᵃᵍᵒ'


def test_humanize_list():
    """Quick tests for humanize.list."""

    assert humanize.list(['one']) == 'one'
    assert humanize.list(['one', 'two']) == 'one and two'
    assert humanize.list(['one', 'two', 'three']) == 'one, two, and three'


def test_humanize_range(monkeypatch):
    """Quick tests for humanize.range (and humanize.date/humanize.time)."""

    def _test(start, end):
        return humanize.range(start, end)

    start = datetime.date(2017, 11, 15)
    start_ts = float(start.strftime('%s'))
    monkeypatch.setattr('time.time', lambda: start_ts)
    next_day = datetime.date(2017, 11, 16)
    next_week = datetime.date(2017, 11, 22)
    next_month = datetime.date(2017, 12, 15)
    next_year = datetime.date(2018, 1, 15)

    assert _test(start, start) == 'Wed 15ᵗʰ'
    assert _test(start, next_day) == 'Wed 15–16ᵗʰ'
    assert _test(start, next_week) == 'Wed 15–22ⁿᵈ'
    assert _test(start, next_month) == 'Wed 15ᵗʰ – Fri, Dec 15ᵗʰ'
    assert _test(start, next_year) == 'Wed 15ᵗʰ – Mon, Jan (2018) 15ᵗʰ'

    start = datetime.datetime(2017, 11, 15, 6)
    start_ts = float(start.strftime('%s'))
    monkeypatch.setattr('time.time', lambda: start_ts)
    next_min = datetime.datetime(2017, 11, 15, 6, 1)
    next_hour = datetime.datetime(2017, 11, 15, 7)
    next_pm = datetime.datetime(2017, 11, 15, 18)
    hours_23 = datetime.datetime(2017, 11, 16, 5)
    hours_25 = datetime.datetime(2017, 11, 16, 7)
    next_month = datetime.datetime(2017, 12, 15, 6)
    next_year = datetime.datetime(2018, 1, 15, 6)

    assert _test(start, next_min) == 'Wed 15ᵗʰ, 6–6:01ᵃᵐ'
    assert _test(start, next_hour) == 'Wed 15ᵗʰ, 6–7ᵃᵐ'
    assert _test(start, next_pm) == 'Wed 15ᵗʰ, 6ᵃᵐ – 6ᵖᵐ'
    assert _test(start, hours_23) == 'Wed 15ᵗʰ, 6ᵃᵐ – 5ᵃᵐ'
    assert _test(start, hours_25) == 'Wed 15ᵗʰ, 6ᵃᵐ – Thu 16ᵗʰ, 7ᵃᵐ'
    assert _test(start, next_month) == 'Wed 15ᵗʰ, 6ᵃᵐ – Fri, Dec 15ᵗʰ, 6ᵃᵐ'
    assert _test(start, next_year) == 'Wed 15ᵗʰ, 6ᵃᵐ – Mon, Jan (2018) 15ᵗʰ, 6ᵃᵐ'



"""Simple container for bot, group, and user configuration."""


class DataSettings:  # pylint: disable=too-few-public-methods
    """Simple container for bot, group, and user configuration."""

    def __init__(self, data, settings):
        self.data = _Container(data)
        self.settings = _Container(settings)


class _Container:

    def __init__(self, data):
        self._data = data

    def __getitem__(self, field):
        return self._data[field]

    def __getattr__(self, field):
        return self[field]

    def __setitem__(self, field, value):
        self._data[field] = value

    def __setattr__(self, field, value):
        if field.startswith('_'):
            return super().__setattr__(field, value)

        self[field] = value



"""Super-simple implementation of https://developers.google.com/protocol-buffers/docs/encoding."""

import functools


class ProtoBuf:
    """Base class for all protobuf files and nested fields."""

    _fields = ()

    def __init__(self, data=None):
        self._setters = {}
        for number, name, subtype, *desc in self._fields:
            if subtype is str:
                decoder = lambda data: bytes(data).decode('utf8')
            elif issubclass(subtype, ProtoBuf):
                decoder = functools.partial(lambda subtype, data: subtype(data), subtype)
            else:
                decoder = lambda data: data
            if desc and desc[0] is list:
                setattr(self, name, [])
                setter = functools.partial(
                    lambda name, decoder, data: getattr(self, name).append(decoder(data)), name,
                    decoder)
            else:
                setattr(self, name, None)
                setter = functools.partial(
                    lambda name, decoder, data: setattr(self, name, decoder(data)), name, decoder)
            self._setters[number] = setter

        if data is not None:
            self.merge_from_bytes(data)

    def merge_from_bytes(self, data):
        """Decode data, an iterable of ints, adding encoded values to self."""

        while data:
            field_type, value = get_field(data)
            if field_type is None:
                break
            if field_type in self._setters:
                self._setters[field_type](value)

    def __iter__(self):
        for _, name, *_ in self._fields:
            yield getattr(self, name)


def get_varint(data):
    """Decode a simple varint."""

    val = exp = 0
    while data:
        chunk = data.pop(0)
        val += (chunk & 0x7f) << (7 * exp)
        if not chunk & 0x80:
            break
        exp += 1
    return val


def get_field(data):
    """Decode a type/length/value tuple."""

    key = get_varint(data)
    wire_type, field_number = key & 0x7, key >> 3
    if wire_type == 0:
        return field_number, get_varint(data)
    if wire_type == 2:
        field_len = get_varint(data)
        value = data[:field_len]
        for _ in range(field_len):
            data.pop(0)
        return field_number, value
    return None, key



"""Tests for metabot.multibot."""

import json
import threading

import ntelebot
import pytest
import yaml

from metabot import multibot


def test_add_bot():
    """Verify the basic behavior of add_bot."""

    mybot = multibot.MultiBot(())
    mockbot = ntelebot.bot.Bot('1234:badbot')
    mockbot.getme.respond(json={'description': 'Unauthorized', 'error_code': 401, 'ok': False})
    with pytest.raises(ntelebot.errors.Unauthorized):
        mybot.add_bot('1234:badbot')


def test_save_load(tmpdir):
    """Verify MultiBot can start with no config, can have a bot added, and can restart."""

    conffile = tmpdir.join('bots.yaml')

    mybot = multibot.MultiBot((), confdir=tmpdir.strpath)
    mockbot = ntelebot.bot.Bot('1234:goodbot')
    mockbot.getme.respond(json={'ok': True, 'result': {'id': 1234, 'username': 'goodbot'}})
    mockbot.getupdates.respond(json={'ok': True, 'result': []})
    mybot.add_bot('1234:goodbot')
    mybot.run_bot('goodbot')
    assert yaml.safe_load(conffile.read()) == {
        'goodbot': {
            'issue37': {
                'telegram': {
                    'running': True,
                    'token': '1234:goodbot',
                },
            },
        },
    }

    tmpdir.join('calendars.json').write(
        json.dumps([
            {
                'calid': 'static:test_multibot',
                'name': 'MultiBot Test Calendar',
            },
        ]))

    newbot = multibot.MultiBot((), confdir=tmpdir.strpath)
    assert newbot.conf == mybot.conf
    assert newbot.calendars == {
        'c9328778': {
            'calid': 'static:test_multibot',
            'name': 'MultiBot Test Calendar',
        },
    }

    mybot.stop_bot('goodbot')
    newbot.stop_bot('goodbot')


def test_module():
    """Verify the configurable module dispatcher."""

    results = []
    event = threading.Event()

    class _DummyMod:  # pylint: disable=too-few-public-methods

        @staticmethod
        def moddispatch(ctx, unused_msg, unused_modconf):  # pylint: disable=missing-docstring
            results.append(ctx.text)
            mybot.stop()
            event.set()

    mybot = multibot.MultiBot({_DummyMod})
    mockbot = ntelebot.bot.Bot('1234:modbot')
    mockbot.getme.respond(json={'ok': True, 'result': {'id': 1234, 'username': 'modbot'}})
    user = {'id': 1000}
    chat = {'id': 1000, 'type': 'private'}
    message = {'message_id': 2000, 'chat': chat, 'from': user, 'text': '/dummymod test'}
    update = {'message': message, 'update_id': 3000}
    mockbot.getupdates.respond(json={'ok': True, 'result': [update]})
    mybot.add_bot('1234:modbot')
    mybot.run_bot('modbot')
    mybot.run()
    event.wait()
    assert results == ['test']



"""Simplified interface to https://docs.python.org/3/library/json.html."""

import json


def load(fname):
    """Load fname as a JSON file, silently returning None on any error."""

    try:
        data = open(fname).read()
    except IOError:
        return

    try:
        return json.loads(data)
    except json.decoder.JSONDecodeError:
        pass


def dump(fname, obj):
    """Save obj as a JSON file to fname."""

    data = json.dumps(obj, indent=4, sort_keys=True).encode('ascii')
    with open(fname, 'wb') as fobj:
        fobj.write(data)
    return obj



"""Tests for metabot.modules.events."""

import pytest

from metabot.calendars import loader
from metabot.modules import events
from metabot.modules import moderator


@pytest.fixture
def conversation(build_conversation, monkeypatch):  # pylint: disable=missing-docstring
    monkeypatch.setattr('time.time', lambda: 2000.)

    cal = loader.get('static:test_events')
    assert cal.calcode == '6fc2c510'
    cal.events = {
        '6fc2c510:alpha': {
            'description': 'Alpha Description',
            'end': 2000,
            'local_id': '6fc2c510:alpha',
            'location': 'Alpha Venue, Rest of Alpha Location',
            'start': 1000,
            'summary': 'Alpha Summary',
        },
        '6fc2c510:bravo': {
            'description': 'Bravo Description',
            'end': 60 * 60 * 24 * 7 + 60 * 60,
            'local_id': '6fc2c510:bravo',
            'location': 'Bravo Venue, Rest of Bravo Location',
            'start': 60 * 60 * 24 * 7,
            'summary': 'Bravo Summary',
        },
        '6fc2c510:charlie': {
            'description': 'Charlie Description',
            'end': 60 * 60 * 24 * 7 + 1 + 60 * 60,
            'local_id': '6fc2c510:charlie',
            'location': 'Charlie Venue, Rest of Charlie Location',
            'start': 60 * 60 * 24 * 7 + 1,
            'summary': 'Charlie Summary',
        },
    }

    conv = build_conversation(events, moderator)
    conv.multibot.multical.add('static:test_events')
    conv.multibot.calendars['6fc2c510'] = {'name': 'Test Calendar'}
    return conv


# pylint: disable=line-too-long


def test_group(conversation, monkeypatch):  # pylint: disable=redefined-outer-name
    """Test /events in a group chat."""

    assert conversation.message('/notevents') == ''

    assert conversation.message('/events', chat_type='supergroup') == """\
[chat_id=-1001000001000 disable_web_page_preview=True parse_mode=HTML reply_to_message_id=2000]
I'm not configured for this group! Ask a bot admin to go into the <b>moderator</b> module settings, group <b>-1001000001000</b>, and choose one or more calendars and set the time zone.
"""

    assert conversation.message(
        '/admin modulestestbot moderator -1001000001000 calendars add 6fc2c510') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › moderator › -1001000001000 › calendars: <b>Select a calendar</b>

Added <code>6fc2c510</code> to your calendar view.

Which calendars should be listed in /events?

Select a calendar to add or remove from the list below:
[Remove Test Calendar | /admin modulestestbot moderator -1001000001000 calendars remove 6fc2c510]
[Back | /admin modulestestbot moderator -1001000001000]
"""

    conversation.bot.config['issue37']['moderator']['-1001000001000']['timezone'] = 'US/Pacific'

    assert conversation.message('/events', chat_type='supergroup') == """\
[chat_id=-1001000001000 disable_web_page_preview=True parse_mode=HTML reply_to_message_id=2000]
<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVUy9QYWNpZmlj">⭐ ᴺᴼᵂ Wed 31ˢᵗ, 4:16–4:33ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
"""

    conversation.bot.config['issue37']['moderator']['-1001000001000']['timezone'] = 'UTC'

    assert conversation.message('/events', chat_type='supergroup') == """\
[chat_id=-1001000001000 disable_web_page_preview=True parse_mode=HTML reply_to_message_id=2000]
<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">⭐ ᴺᴼᵂ Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVVEM">¹ʷ Thu 8ᵗʰ, 12–1ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>
"""

    conversation.bot.config['issue37']['moderator']['-1001000001000']['maxeventscount'] = 1

    assert conversation.message('/events', chat_type='supergroup') == """\
[chat_id=-1001000001000 disable_web_page_preview=True parse_mode=HTML reply_to_message_id=2000]
<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVVEM">⭐ ᴺᴼᵂ Thu 1ˢᵗ, 12:16–12:33ᵃᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>
"""

    monkeypatch.setattr('time.time', lambda: 2000000.)

    assert conversation.message('/events', chat_type='supergroup') == """\
[chat_id=-1001000001000 disable_web_page_preview=True parse_mode=HTML reply_to_message_id=2000]
No events in the next 6 days!
"""


def test_private(conversation, monkeypatch):  # pylint: disable=redefined-outer-name
    """Test /events in a private chat."""

    assert conversation.message('/events', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Please choose one or more calendars and set your time zone!
[Settings | /events set]
"""

    assert conversation.message('/events set', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings: <b>Choose a field</b>
[calendars • Which calendars do you want to see? | /events set calendars]
[timezone • What time zone are you in? | /events set timezone]
[Back | /events]
"""

    assert conversation.message('/events set calendars', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Add Test Calendar | /events set calendars add 6fc2c510]
[Back | /events set]
"""

    assert conversation.message('/events set calendars add 6fc2c510', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

Added <code>6fc2c510</code> to your calendar view.

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Remove Test Calendar | /events set calendars remove 6fc2c510]
[Back | /events set]
"""

    conversation.bot.config['issue37']['events']['users']['2000']['timezone'] = 'US/Pacific'

    assert conversation.message('/events', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVUy9QYWNpZmlj">⭐ ᴺᴼᵂ Wed 31ˢᵗ, 4:16–4:33ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>

Alpha Description
[\xa0 | /stop] [Settings | /events set] [Next | /events 6fc2c510:bravo]
"""

    assert conversation.message('/events 6fc2c510:bravo', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
<b>Bravo Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVUy9QYWNpZmlj">¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>

Bravo Description
[Prev | /events 6fc2c510:alpha] [Current | /events] [Next | /events 6fc2c510:charlie]
"""

    monkeypatch.setattr('time.time', lambda: 2000000.)

    assert conversation.message('/events', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
No upcoming events!
[\xa0 | /stop] [Settings | /events set] [\xa0 | /stop]
"""


def test_inline(conversation, monkeypatch):  # pylint: disable=redefined-outer-name
    """Test @BOTUSER events."""

    assert conversation.raw_inline('events') == [
        {
            'cache_time': 30,
            'inline_query_id': 2000,
            'is_personal': True,
            'switch_pm_text': 'Click to choose one or more calendars and set your time zone!',
            'switch_pm_parameter': 'L2V2ZW50cw',
            'results': [],
        },
    ]  # yapf: disable

    assert conversation.message('/events set calendars add 6fc2c510') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

Added <code>6fc2c510</code> to your calendar view.

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Remove Test Calendar | /events set calendars remove 6fc2c510]
[Back | /events set]
"""

    conversation.bot.config['issue37']['events']['users']['1000']['timezone'] = 'US/Pacific'

    assert conversation.raw_inline('events') == [
        {
            'cache_time': 30,
            'inline_query_id': 2000,
            'is_personal': True,
            'switch_pm_text': 'Settings',
            'switch_pm_parameter': 'L2V2ZW50cyBzZXQ',
            'results': [
                {
                    'description': '⭐ ᴺᴼᵂ Wed 31ˢᵗ, 4:16–4:33ᵖᵐ @ Alpha Venue',
                    'id': '6fc2c510:alpha',
                    'input_message_content': {
                        'disable_web_page_preview': True,
                        'message_text': '<b>Alpha Summary</b>\n'
                                        '<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVUy9QYWNpZmlj">⭐ ᴺᴼᵂ Wed 31ˢᵗ, 4:16–4:33ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>',
                        'parse_mode': 'HTML',
                    },
                    'title': 'Alpha Summary',
                    'type': 'article',
                },
                {
                    'description': '¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ @ Bravo Venue',
                    'id': '6fc2c510:bravo',
                    'input_message_content': {
                        'disable_web_page_preview': True,
                        'message_text': '<b>Bravo Summary</b>\n'
                                        '<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVUy9QYWNpZmlj">¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>',
                        'parse_mode': 'HTML',
                    },
                    'title': 'Bravo Summary',
                    'type': 'article',
                },
                {
                    'description': '¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ @ Charlie Venue',
                    'id': '6fc2c510:charlie',
                    'input_message_content': {
                        'disable_web_page_preview': True,
                        'message_text': '<b>Charlie Summary</b>\n'
                                        '<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpjaGFybGllIFVTL1BhY2lmaWM">¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Charlie+Venue%2C+Rest+of+Charlie+Location">Charlie Venue</a>',
                        'parse_mode': 'HTML',
                    },
                    'title': 'Charlie Summary',
                    'type': 'article',
                },
            ],
        },
    ]  # yapf: disable

    assert conversation.raw_inline('events bra') == [
        {
            'cache_time': 30,
            'inline_query_id': 2000,
            'is_personal': True,
            'switch_pm_text': 'Settings',
            'switch_pm_parameter': 'L2V2ZW50cyBzZXQ',
            'results': [
                {
                    'description': '¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ @ Bravo Venue',
                    'id': '6fc2c510:bravo',
                    'input_message_content': {
                        'disable_web_page_preview': True,
                        'message_text': '<b>Bravo Summary</b>\n'
                                        '<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVUy9QYWNpZmlj">¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>',
                        'parse_mode': 'HTML',
                    },
                    'title': 'Bravo Summary',
                    'type': 'article',
                },
            ],
        },
    ]  # yapf: disable

    assert conversation.raw_inline('events full bra') == [
        {
            'cache_time': 30,
            'inline_query_id': 2000,
            'is_personal': True,
            'switch_pm_text': 'Settings',
            'switch_pm_parameter': 'L2V2ZW50cyBzZXQ',
            'results': [
                {
                    'description': 'Bravo Description',
                    'id': '6fc2c510:bravo',
                    'input_message_content': {
                        'disable_web_page_preview': True,
                        'message_text': '<b>Bravo Summary</b>\n'
                                        '<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDpicmF2byBVUy9QYWNpZmlj">¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Bravo+Venue%2C+Rest+of+Bravo+Location">Bravo Venue</a>\n'
                                        '\n'
                                        'Bravo Description',
                        'parse_mode': 'HTML',
                    },
                    'title': 'Bravo Summary • ¹ʷ Wed, Jan (1970) 7ᵗʰ, 4–5ᵖᵐ @ Bravo Venue',
                    'type': 'article',
                },
            ],
        },
    ]  # yapf: disable

    monkeypatch.setattr('time.time', lambda: 2000000.)

    assert conversation.raw_inline('events') == [
        {
            'cache_time': 30,
            'inline_query_id': 2000,
            'is_personal': True,
            'switch_pm_text': 'Settings',
            'switch_pm_parameter': 'L2V2ZW50cyBzZXQ',
            'results': [],
        },
    ]  # yapf: disable


def test_customize(conversation):  # pylint: disable=redefined-outer-name
    """Test /events admin."""

    conversation.bot.config['issue37']['events']['users']['1000']['calendars'] = '6fc2c510'
    conversation.bot.config['issue37']['events']['users']['1000']['timezone'] = 'US/Pacific'

    assert conversation.message('/events') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
<b>Alpha Summary</b>
<a href="https://t.me/modulestestbot?start=L2V2ZW50cyA2ZmMyYzUxMDphbHBoYSBVUy9QYWNpZmlj">⭐ ᴺᴼᵂ Wed 31ˢᵗ, 4:16–4:33ᵖᵐ</a> @ <a href="https://maps.google.com/maps?q=Alpha+Venue%2C+Rest+of+Alpha+Location">Alpha Venue</a>

Alpha Description
[Customize | /events admin 6fc2c510:alpha]
[\xa0 | /stop] [Settings | /events set] [Next | /events 6fc2c510:bravo]
"""

    assert conversation.message('/events admin 6fc2c510:alpha') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Customize › 6fc2c510:alpha: <b>Choose a field</b>
[event • What banner image should be used for this specific event? | /events admin 6fc2c510:alpha event]
[series • What banner image should be used for events with this title? | /events admin 6fc2c510:alpha series]
[Back | /events admin]
"""

    assert conversation.message('/events admin 6fc2c510:alpha series') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Customize › 6fc2c510:alpha › series: <b>Send a new image for Alpha Summary</b>

Send a new image, or type "off" to disable/reset to default.
[Back | /events admin 6fc2c510:alpha]
"""

    assert conversation.message('https://SERIES-IMAGE') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Customize › 6fc2c510:alpha › series

Set <code>Alpha Summary</code> to <code>https://SERIES-IMAGE</code>.
[Back | /events admin 6fc2c510:alpha]
"""

    assert conversation.bot.config['issue37']['events']['series'] == {
        'Alpha Summary': 'https://SERIES-IMAGE',
    }

    assert conversation.message('/events admin 6fc2c510:alpha event') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Customize › 6fc2c510:alpha › event: <b>Send a new image for 6fc2c510:alpha</b>

Send a new image, or type "off" to disable/reset to default.
[Back | /events admin 6fc2c510:alpha]
"""

    assert conversation.message('https://EVENT-ICON') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Customize › 6fc2c510:alpha › event

Set <code>6fc2c510:alpha</code> to <code>https://EVENT-ICON</code>.
[Back | /events admin 6fc2c510:alpha]
"""

    assert conversation.bot.config['issue37']['events']['events'] == {
        '6fc2c510:alpha': 'https://EVENT-ICON',
    }


def test_settings(conversation):  # pylint: disable=redefined-outer-name
    """Test /events set."""

    assert conversation.message('/events set') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings: <b>Choose a field</b>
[calendars • Which calendars do you want to see? | /events set calendars]
[timezone • What time zone are you in? | /events set timezone]
[Back | /events]
"""

    assert conversation.message('/events set calendars') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Add Test Calendar | /events set calendars add 6fc2c510]
[Back | /events set]
"""

    assert conversation.message('/events set calendars add 6fc2c510') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

Added <code>6fc2c510</code> to your calendar view.

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Remove Test Calendar | /events set calendars remove 6fc2c510]
[Back | /events set]
"""

    assert conversation.message('/events set calendars add 6fc2c510') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

<code>6fc2c510</code> is already in your calendar view!

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Remove Test Calendar | /events set calendars remove 6fc2c510]
[Back | /events set]
"""

    assert conversation.message('/events set calendars remove 6fc2c510') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

Removed <code>6fc2c510</code> from your calendar view.

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Add Test Calendar | /events set calendars add 6fc2c510]
[Back | /events set]
"""

    assert conversation.message('/events set calendars remove 6fc2c510') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

<code>6fc2c510</code> is not in your calendar view!

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Add Test Calendar | /events set calendars add 6fc2c510]
[Back | /events set]
"""

    assert conversation.message('/events set calendars add bogus') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings › calendars: <b>Select a calendar</b>

<code>bogus</code> is not a calendar!

Which calendars do you want to see?

Select a calendar to add or remove from the list below:
[Add Test Calendar | /events set calendars add 6fc2c510]
[Back | /events set]
"""

    assert conversation.message('/events set bogus') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Events › Settings: <b>Choose a field</b>

I can't set <code>bogus</code>.
[calendars • Which calendars do you want to see? | /events set calendars]
[timezone • What time zone are you in? | /events set timezone]
[Back | /events]
"""


def test_help(conversation):  # pylint: disable=redefined-outer-name
    """Test /help."""

    assert conversation.message('/help', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
<b>Commands</b>

/events – Display recent and upcoming events
"""



"""Tests for metabot.modules.telegram."""

import pytest

from metabot.modules import telegram


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(telegram)


# pylint: disable=line-too-long


def test_admin(conversation):  # pylint: disable=redefined-outer-name
    """Test /admin BOTNAME telegram."""

    assert conversation.message('/admin modulestestbot telegram') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › telegram: <b>Choose an action</b>
[Start bot | /admin modulestestbot telegram start]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot telegram start') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › telegram: <b>Choose an action</b>

@modulestestbot is now running.
[Stop bot | /admin modulestestbot telegram stop]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot telegram start') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › telegram: <b>Choose an action</b>

@modulestestbot is already running.
[Stop bot | /admin modulestestbot telegram stop]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot telegram stop') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › telegram: <b>Choose an action</b>

@modulestestbot is now offline.
[Start bot | /admin modulestestbot telegram start]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot telegram stop') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › telegram: <b>Choose an action</b>

@modulestestbot is not currently running.
[Start bot | /admin modulestestbot telegram start]
[Back | /admin modulestestbot]
"""



"""Tests for metabot.modules.groups."""

import pytest

from metabot.modules import groups


@pytest.fixture
def conversation(build_conversation):  # pylint: disable=missing-docstring
    return build_conversation(groups)


# pylint: disable=line-too-long


def test_conversation(conversation, requests_mock):  # pylint: disable=redefined-outer-name
    """Test /groups, @BOTNAME groups, and /admin BOTNAME groups."""

    assert conversation.message('/groups') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Group List

I don't know about any public groups yet, sorry!
"""

    assert conversation.message('/admin modulestestbot groups') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › groups: <b>Type a group username or invite link</b>

Type the <code>@USERNAME</code> or <code>https://t.me/joinchat/INVITE_LINK</code> for the group you want to add, or select an existing group to remove:
[Back | /admin modulestestbot]
"""

    conversation.bot.get_chat.respond(
        json={
            'ok': True,
            'result': {
                'id': -1001000001000,
                'title': 'Dummy Public Group!',
                'username': 'DummyGroup',
                'type': 'supergroup',
                'description': 'Dummy public group'
            }
        })
    conversation.bot.get_chat_administrators.respond(json={
        'ok': True,
        'result': [{
            'user': {
                'id': 1000,
                'username': 'DummyUser',
            },
        }],
    })

    assert conversation.message('@dummygroup') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › groups: <b>Type a group username or invite link</b>

Added <b>Dummy Public Group!</b>.

Type the <code>@USERNAME</code> or <code>https://t.me/joinchat/INVITE_LINK</code> for the group you want to add, or select an existing group to remove:
[Dummy Public Group! • DummyGroup | /admin modulestestbot groups remove 92aa63]
[Back | /admin modulestestbot]
"""

    requests_mock.get('https://t.me/joinchat/DUMMY_INVITE_LINK',
                      text="""
<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8">
    <title>Telegram: Join Group Chat</title>
    <meta name="viewport" content="width=device-width, initial-scale=1.0">

<meta property="og:title" content="Dummy Private Group&#33;">
<meta property="og:image" content="https://cdn1.telesco.pe/file/DUMMY_PATH.jpg">
<meta property="og:site_name" content="Telegram">
<meta property="og:description" content="Dummy private group">
""")

    assert conversation.message('https://t.me/joinchat/DUMMY_INVITE_LINK') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › groups: <b>Type a group username or invite link</b>

Added <b>Dummy Private Group!</b>.

Type the <code>@USERNAME</code> or <code>https://t.me/joinchat/INVITE_LINK</code> for the group you want to add, or select an existing group to remove:
[Dummy Private Group! • https://t.me/joinchat/DUMMY_INVITE_LINK | /admin modulestestbot groups remove c5bfc8]
[Dummy Public Group! • DummyGroup | /admin modulestestbot groups remove 92aa63]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/groups') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Group List: <b>Choose a location</b>
[Worldwide | /groups Worldwide]
"""

    assert conversation.message('/groups Worldwide') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Group List › Worldwide

<a href="https://t.me/joinchat/DUMMY_INVITE_LINK">Dummy Private Group!</a>
Dummy private group

<a href="https://t.me/DummyGroup">Dummy Public Group!</a>
Dummy public group
[Back | /groups]
"""

    assert conversation.raw_inline('') == []

    assert conversation.raw_inline('groups') == [
        {
            'cache_time': 60,
            'inline_query_id': 2000,
            'results': [
                {
                    'description': 'Dummy private group',
                    'id': 'https://t.me/joinchat/DUMMY_INVITE_LINK',
                    'input_message_content': {
                        'disable_web_page_preview': True,
                        'message_text': '<a href="https://t.me/joinchat/DUMMY_INVITE_LINK">Dummy Private Group!</a>\n'
                                        'Dummy private group',
                        'parse_mode': 'HTML',
                    },
                    'title': 'Dummy Private Group! • https://t.me/joinchat/DUMMY_INVITE_LINK',
                    'type': 'article',
                },
                {
                    'description': 'Dummy public group',
                    'id': 'https://t.me/DummyGroup',
                    'input_message_content': {
                        'disable_web_page_preview': True,
                        'message_text': '<a href="https://t.me/DummyGroup">Dummy Public Group!</a>\n'
                                        'Dummy public group',
                        'parse_mode': 'HTML',
                    },
                    'title': 'Dummy Public Group! • DummyGroup',
                    'type': 'article',
                },
            ],
        }
    ]  # yapf: disable

    assert conversation.raw_inline('groups priva') == [
        {
            'cache_time': 60,
            'inline_query_id': 2000,
            'results': [
                {
                    'description': 'Dummy private group',
                    'id': 'https://t.me/joinchat/DUMMY_INVITE_LINK',
                    'input_message_content': {
                        'disable_web_page_preview': True,
                        'message_text': '<a href="https://t.me/joinchat/DUMMY_INVITE_LINK">Dummy Private Group!</a>\n'
                                        'Dummy private group',
                        'parse_mode': 'HTML',
                    },
                    'title': 'Dummy Private Group! • https://t.me/joinchat/DUMMY_INVITE_LINK',
                    'type': 'article',
                },
            ],
        }
    ]  # yapf: disable

    assert conversation.message('/admin modulestestbot groups remove blah') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › groups: <b>Type a group username or invite link</b>

Oops, the groups list changed since you loaded that screen, try again.

Type the <code>@USERNAME</code> or <code>https://t.me/joinchat/INVITE_LINK</code> for the group you want to add, or select an existing group to remove:
[Dummy Private Group! • https://t.me/joinchat/DUMMY_INVITE_LINK | /admin modulestestbot groups remove c5bfc8]
[Dummy Public Group! • DummyGroup | /admin modulestestbot groups remove 92aa63]
[Back | /admin modulestestbot]
"""

    assert conversation.message('/admin modulestestbot groups remove 92aa63') == """\
[chat_id=1000 disable_web_page_preview=True parse_mode=HTML]
Bot Admin › modulestestbot › groups: <b>Type a group username or invite link</b>

Removed <b>Dummy Public Group!</b>.

Type the <code>@USERNAME</code> or <code>https://t.me/joinchat/INVITE_LINK</code> for the group you want to add, or select an existing group to remove:
[Dummy Private Group! • https://t.me/joinchat/DUMMY_INVITE_LINK | /admin modulestestbot groups remove c5bfc8]
[Back | /admin modulestestbot]
"""


def test_help(conversation):  # pylint: disable=redefined-outer-name
    """Test /help."""

    assert conversation.message('/help', user_id=2000) == """\
[chat_id=2000 disable_web_page_preview=True parse_mode=HTML]
<b>Commands</b>

/groups – Find other group chats
"""



"""Quick routine to extract ticket information from event descriptions."""

import json
import logging
import re

import ntelebot


def get_info(text):
    """Quick routine to extract ticket information from event descriptions."""

    for provider in (_brownpapertickets, _eventbrite):
        yield from provider(text)


def _brownpapertickets(text):
    events = {
        event_id for _, _, event_id in re.findall(
            '://(www.|m.|)(brownpapertickets.com|bpt.me)/event/([0-9]+)', text)
    }
    for event_id in sorted(events):
        ret = re.search(
            '>([0-9]+) tickets remaining',
            ntelebot.requests.get(f'https://m.bpt.me/event/{event_id}', timeout=10).text)
        if ret:
            remaining = int(ret.groups()[0])
        else:
            remaining = 0
        yield remaining, f'https://bpt.me/event/{event_id}'


def _eventbrite(text):
    events = {
        event_id for _, event_id in re.findall('://www.eventbrite.com/e/([^/]+-)?([0-9]+)', text)
    }
    for event_id in sorted(events):
        url = f'https://www.eventbrite.com/ajax/event/{event_id}/ticket_classes/for_sale/'
        try:
            data = ntelebot.requests.get(url,
                                         timeout=10,
                                         params={
                                             'pos': 'online',
                                         },
                                         headers={
                                             'x-requested-with': 'XMLHttpRequest',
                                         }).json()
        except json.decoder.JSONDecodeError:
            logging.exception('Decoding %s:', url)
        else:
            remaining = 0
            for ticket_class in data['ticket_classes']:
                if ticket_class['number_of_tickets_remaining']:  # If unset, this is None, not 0.
                    remaining += ticket_class['number_of_tickets_remaining']
            yield remaining, data['event_info']['vanity_url']



"""Display recent and upcoming events."""

import datetime
import logging
import time
import urllib.parse

import pytz

from metabot.util import geoutil
from metabot.util import html
from metabot.util import humanize
from metabot.util import icons
from metabot.util import tickets

# https://www.weather.gov/forecast-icons
BAD_WEATHER_KEYWORDS = {
    'blizzard', 'freezing', 'heavy', 'hurricane', 'ice', 'snow', 'tornado', 'tropical'
}


class CalendarConf:  # pylint: disable=too-few-public-methods
    """A group's calendar configuration."""

    def __init__(self, groupconf):
        self.calcodes = groupconf.get('calendars', '').split()
        self.count = groupconf.get('maxeventscount', 10)
        self.days = groupconf.get('maxeventsdays', 6)
        timezone = groupconf.get('timezone')
        self.tzinfo = timezone and pytz.timezone(timezone)

    def get_events(self, bot, *, when=None):
        """Build lists of events for the given time (or now)."""

        return _get_group_events(bot, self.calcodes, self.tzinfo, self.count, self.days, now=when)


def _get_group_events(bot, calcodes, tzinfo, count, days, *, now=None):  # pylint: disable=too-many-arguments,too-many-locals
    """Build lists of events for the given calendar configuration."""

    calendar_view = bot.multibot.multical.view(calcodes)
    if now is None:
        now = time.time()
    nowdt = datetime.datetime.fromtimestamp(now, tzinfo)
    midnight = nowdt.replace(hour=0, minute=0, second=0, microsecond=0)
    period = (midnight + datetime.timedelta(days=days + 1) - nowdt).total_seconds()
    return list(calendar_view.get_overlap(now, now + period))[:count]


def format_event(bot, event, tzinfo, *, full=True, base=None, countdown=True):  # pylint: disable=too-many-arguments
    """Given a metabot.calendars.base.Calendar event, build a human-friendly representation."""

    message = '<b>%s</b>' % html.escape(event['summary'])
    for count, url in tickets.get_info(event['description']):
        if count:
            message = '%s [<a href="%s">%i tickets remaining</a>]' % (message, url, count)
        else:
            message = '%s [<a href="%s">tickets</a>]' % (message, url)
    message = '%s\n<a href="%s">%s</a>' % (
        message, bot.encode_url('/events %s %s' % (event['local_id'], tzinfo.zone)),
        humanize_range(event['start'], event['end'], tzinfo, base=base, countdown=countdown))
    if event['location']:
        location_name = event['location'].split(',', 1)[0]
        location_url = 'https://maps.google.com/maps?' + urllib.parse.urlencode({
            'q': event['location'].encode('utf-8'),
        })  # yapf: disable
        message = '%s @ <a href="%s">%s</a>' % (message, html.escape(location_url),
                                                html.escape(location_name))
        geo = format_geo(event['location'], event['start'])
        if geo:
            message = '%s\n\u26a0 %s' % (message, geo)
    if full and event['description']:
        message = '%s\n\n%s' % (message, html.sanitize(event['description']))
    return message


def format_events(bot, events, tzinfo, *, base=None, countdown=True):
    """Prepare a message containing human-friendly representations of the given events."""

    return '\n'.join(
        format_event(bot, event, tzinfo, full=False, base=base, countdown=countdown)
        for event in events)


def format_geo(address, now):
    """Build a string of weather exceptions for the given address as of the given time."""

    try:
        forecast = geoutil.hourlyforecast(address, now)
    except Exception:  # pylint: disable=broad-except
        logging.exception('Ignoring:')
        return
    if not forecast:
        return
    warnings = []
    if set(forecast['shortForecast'].lower().split()) & BAD_WEATHER_KEYWORDS:
        warnings.append(forecast['shortForecast'])
    if warnings:
        if forecast['temperatureUnit'] == 'F':
            warnings.append('%s\u2109' % forecast['temperature'])
        elif forecast['temperatureUnit'] == 'C':
            warnings.append('%s\u2103' % forecast['temperature'])
        return ' \u2022 '.join(warnings)


def humanize_range(start, end, tzinfo, *, base=None, countdown=True):
    """Return the range between start and end as human-friendly text."""

    startdt = datetime.datetime.fromtimestamp(start, tzinfo)
    enddt = datetime.datetime.fromtimestamp(end, tzinfo)
    text = humanize.range(startdt, enddt, base=base)
    if countdown:
        text = f'{humanize.howrecent(startdt, enddt, base=base)} {text}'
    return text


def get_image(event, botconf):
    """Choose the best image for a given event and botconf."""

    eventconf = botconf['issue37']['events']
    eventimage = eventconf['events'].get(event['local_id'])
    if eventimage:
        return eventimage
    for pattern, seriesimage in eventconf['series'].items():
        if pattern.lower() in event['summary'].lower():
            return seriesimage
    return icons.match(event['summary']) or icons.match(event['description'])



"""Tests for metabot.calendars.loader."""

import os
import pickle

from metabot.calendars import loader


def test_loader(monkeypatch, tmpdir):
    """Run through loader.get."""

    # pylint: disable=protected-access
    monkeypatch.setattr(loader._CachingCalendarMixin, '_cache_dir', tmpdir.strpath)
    fname = tmpdir.strpath + '/3ccaceeb.pickle'
    with open(fname, 'wb') as cachefile:
        cachefile.write(pickle.dumps({'test_value': 1}))
    cal = loader.get('static:alpha@example.com')
    assert cal._CachingCalendarMixin__fname == fname
    assert cal.test_value == 1
    os.remove(fname)

    assert not cal.poll()
    assert not os.path.exists(fname)

    cal.poll_result = True
    assert cal.poll()
    assert os.path.exists(fname)
    os.remove(fname)

    local = cal.add({
        'start': 1000,
        'end': 2000,
    })
    assert local
    assert os.path.exists(fname)
    os.remove(fname)

    assert cal.update(local['local_id'], {'summary': 'new summary'})
    assert os.path.exists(fname)
    os.remove(fname)

    cal.remove(local['local_id'])
    assert os.path.exists(fname)
    os.remove(fname)



"""Quick utilities for geocoding an address and looking up its weather forecast."""

import logging
import time
import urllib.parse

import googlemaps
import ntelebot

from metabot.util import iso8601
from metabot.util import pickleutil

try:
    _CLIENT_KEY = next(open('config/google_maps_apikey')).strip()
except IOError:
    _CLIENT = None
else:
    _CLIENT = googlemaps.Client(key=_CLIENT_KEY)
_CACHEFILE = 'config/geoutil.pickle'
_CACHE = pickleutil.load(_CACHEFILE) or {}
_SHORTCACHE = {}


def _save():
    pickleutil.dump(_CACHEFILE, _CACHE)


def geocode(address):
    """Look up the given address in Google Maps."""

    if _CLIENT is None:
        return
    if 'geocode' not in _CACHE:
        _CACHE['geocode'] = {}
    if address not in _CACHE['geocode']:
        logging.info('Geocoding %r.', address)
        _CACHE['geocode'][address] = _CLIENT.geocode(address)
        _save()
    return _CACHE['geocode'][address]


def _weatherfetch(url):
    url = urllib.parse.urljoin('https://api.weather.gov/', url)
    headers = {
        'user-agent': 'https://github.com/nmlorg/metabot',
    }
    return ntelebot.requests.get(url, headers=headers, timeout=10).json()


def _weatherpoint(lat, lon):
    """Map a latitude/longitude to a weather.gov gridpoint."""

    # https://weather-gov.github.io/api/general-faqs#how-do-i-get-a-forecast-for-a-location-from-the-api
    lat = ('%.4f' % lat).rstrip('0')
    lon = ('%.4f' % lon).rstrip('0')
    key = '%s,%s' % (lat, lon)

    if 'weatherpoint' not in _CACHE:
        _CACHE['weatherpoint'] = {}
    if key not in _CACHE['weatherpoint']:
        _CACHE['weatherpoint'][key] = _weatherfetch('points/' + key)
        _save()
    return _CACHE['weatherpoint'][key]


def _hourlyforecast(lat, lon, when):
    """Retrieve the hourly forecast for the given latitude/longitude."""

    point = _weatherpoint(lat, lon)
    if not point.get('properties'):
        return
    now = time.time()
    url = point['properties']['forecastHourly']
    last, ret = _SHORTCACHE.get(url) or (0, None)
    if last < now - 10 * 60:
        ret = _weatherfetch(url)['properties']['periods']
        _SHORTCACHE[url] = (now, ret)

    for period in ret:
        start = iso8601.totimestamp(period['startTime'])
        end = iso8601.totimestamp(period['endTime'])
        if start <= when < end:
            return period


def hourlyforecast(address, when):
    """Retrieve the hourly forecast for the given address."""

    geo = geocode(address)
    if not geo:
        return
    return _hourlyforecast(geo[0]['geometry']['location']['lat'],
                           geo[0]['geometry']['location']['lng'], when)



"""Tests for metabot.util.pickleutil."""

import json

from metabot.util import pickleutil


def test_optimize_simple():
    """Verify the optimizer preserves simple objects."""

    assert pickleutil.optimize(b'str') == b'str'
    assert pickleutil.optimize('str') == 'str'
    assert pickleutil.optimize({b'a': 1}) == {b'a': 1}
    assert pickleutil.optimize([1, 2]) == [1, 2]
    assert pickleutil.optimize((1, 2)) == (1, 2)
    assert pickleutil.optimize(1) == 1


def test_optimize_strings():
    """Verify the optimizer normalizes and dedupes strings."""

    bytes_1 = json.loads('"str test"').encode('ascii')
    bytes_2 = json.loads('"str test"').encode('ascii')
    assert bytes_1 == bytes_2 and bytes_1 is not bytes_2

    unicode_1 = json.loads('"str test"')
    unicode_2 = json.loads('"str test"')
    assert unicode_1 == unicode_2 and unicode_1 is not unicode_2

    store = {}

    optimized_str = pickleutil.optimize(bytes_1, store=store)
    assert pickleutil.optimize(bytes_2, store=store) is optimized_str
    optimized_uni = pickleutil.optimize(unicode_1, store=store)
    assert pickleutil.optimize(unicode_2, store=store) is optimized_uni


def test_optimize_containers():
    """Verify the optimizer doesn't break linkages by recreating containers."""

    key1 = json.loads('"key test"')
    key2 = json.loads('"key test"')
    assert key1 == key2 and key1 is not key2

    entry = {'dummy': 'test'}
    cont = {'dummy1': {key1: entry}, 'dummy2': {key2: entry}}
    (key1copy,) = tuple(cont['dummy1'])
    (key2copy,) = tuple(cont['dummy2'])
    assert key1copy == key2copy and key1copy is not key2copy
    assert cont['dummy1'][key1] is cont['dummy2'][key2]

    new = pickleutil.optimize(cont)
    assert new == cont
    (key1copy,) = tuple(new['dummy1'])
    (key2copy,) = tuple(new['dummy2'])
    assert key1copy is key2copy
    assert new['dummy1'][key1] is new['dummy2'][key2]


def test_load_nonexistent(tmpdir):
    """Verify the loader silently ignores nonexistent files."""

    tmpfile = tmpdir.join('test.pickle')
    assert pickleutil.load(tmpfile.strpath) is None


def test_load_empty(tmpdir):
    """Verify the loader silently ignores empty files."""

    tmpfile = tmpdir.join('test.pickle')
    tmpfile.write('')
    assert pickleutil.load(tmpfile.strpath) is None


def test_load_malformed(tmpdir):
    """Verify the loader silently ignores malformed files."""

    tmpfile = tmpdir.join('test.pickle')
    tmpfile.write('bogus data')
    assert pickleutil.load(tmpfile.strpath) is None


def test_dump_simple(tmpdir):
    """Verify the dumper preserves simple objects."""

    tmpfile = tmpdir.join('test.pickle')
    obj = 1
    assert pickleutil.dump(tmpfile.strpath, obj) == obj
    assert tmpfile.load() == obj


def test_dump_load(tmpdir):
    """Verify the dumper and loader work as expected."""

    tmpfile = tmpdir.join('test.pickle')
    obj = {b'key': [b'value', b'value']}
    assert pickleutil.dump(tmpfile.strpath, obj) == obj
    assert pickleutil.load(tmpfile.strpath) == obj
